/* 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

package android.androidVNC;

import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.zip.DataFormatException;
import java.util.zip.Deflater;
import java.util.zip.Inflater;

import android.util.Log;
import android.widget.Toast;
/** 
 * <p>(Se omiten acentos)</p>
 * <p>FileTransfer.java</p>
 * <p>Autor: Alberto Poblacion Navarro</p>
 * <p>Aadido al proyecto en la version v1.0</p>
 * <p><b> Proyecto original:</b></p>
 		* <p>Web: http://code.google.com/p/android-vnc-viewer/</p>
 		* <p>Release: 39</p>
 		* <p>Migrada a release 60.</p>
 * <p><b>Codigo obtenido de:</b> http://sourceforge.net/projects/ultravnc</p>
 * <p>Aplicacion: Java viewer</p>
 * <p>Archivo RfbProto.java</p>
 * <p>Release: 346</p>
 * 
 * <p>Clase utilizada para menejar un cuadro de dialogo que permitia el intercambio de ficheros. Se aprovechan las funciones para la 
 * transferencia de archivos pero se elimina todo lo referente al cuadro de dialogo original. Para mas informacion ver el archivo original.
 * Ademas se aaden las sentencias necesarias para liberar el semaforo se obtiene la informacion de un directorio remoto 
 * para mostrarse por pantalla. 
 * Se aade tambien las notificaciones al usuario y el registro de eventos.
 * Eliminacion de todo lo referente al cuadro de dialogo de su utilidad original.</p>
*/
public class FileTransfer {
	static String TAG="FileTransfer";
	
	private VncCanvas vncCanvas;
	private VncCanvasActivity vncCanvasActivity;
	static private FileTransfer instancia=null;
	
	static FileTransfer Instancia (VncCanvas vnccanvas, VncCanvasActivity vnccanvasactivity){
		if(instancia == null)
			instancia=new FileTransfer(vnccanvas, vnccanvasactivity);
		return instancia;
	}
	
	OutputStream os;
	DataInputStream is;
	RfbProto rfb;
	AndroidLog log;
	Notificacion notificacion;
	
	private FileTransfer(VncCanvas vnccanvas, VncCanvasActivity vnccanvasactivity) {
		vncCanvas=vnccanvas;
		vncCanvasActivity= vnccanvasactivity;
		os=vncCanvas.rfb.os;
		is=vncCanvas.rfb.is;
		rfb=vnccanvas.rfb;
	
		a = new ArrayList();
		// sf@2004
		remoteDirsList = new ArrayList();
		remoteFilesList = new ArrayList();
		
		log = AndroidLog.Instancia();
		notificacion = Notificacion.Instancia(vncCanvasActivity);
		
	}
	// sf@2004 - FileTransfer part
	ArrayList remoteDirsList;
	ArrayList remoteFilesList;
	ArrayList a;
	//boolean fFTInit = true; // sf@2004
	//boolean fFTAllowed = true;
	boolean fAbort = false;
	boolean fFileReceptionError = false;
	boolean fFileReceptionRunning = false;
	boolean inDirectory2;
	FileOutputStream fos;
	FileInputStream fis;
	String sendFileSource; //Cadena donde esta el path del archivo a enviar
	String receivePath; //cadena del path recibida
	long fileSize;
	long receiveFileSize;
	long fileChunkCounter;

	
	final static int rfbFileTransfer = 7;
	
	final static int sz_rfbFileTransferMsg = 12,
	// FileTransfer Content types and Params defines
	rfbDirContentRequest = 1,
	// Client asks for the content of a given Server directory
	rfbDirPacket = 2, // Full directory name or full file name.
	// Null content means end of Directory
	rfbFileTransferRequest = 3,
	// Client asks the server for the tranfer of a given file
	rfbFileHeader = 4,
	// First packet of a file transfer, containing file's features
	rfbFilePacket = 5, // One slice of the file
	rfbEndOfFile = 6,// End of file transfer (the file has been received or error)
	rfbAbortFileTransfer = 7, // The file transfer must be aborted, whatever the state
	rfbFileTransferOffer = 8,
	// The client offers to send a file to the server
	rfbFileAcceptHeader = 9, // The server accepts or rejects the file
	rfbCommand = 10,
	// The Client sends a simple command (File Delete, Dir create etc...)
	rfbCommandReturn = 11,
	//	New FT Protocole (V2) The zipped checksums of the destination file (Delta Transfer)
	rfbFileChecksums = 12,
	// The Client receives the server's answer about a simple command
	// rfbDirContentRequest client Request - content params 
	rfbRDirContent = 1, // Request a Server Directory contents
	rfbRDrivesList = 2, // Request the server's drives list
	
	// rfbDirPacket & rfbCommandReturn  server Answer - content params
	rfbADirectory = 1, // Reception of a directory name
	//rfbAFile = 2, // Reception of a file name 
	rfbADrivesList = 3, // Reception of a list of drives
	rfbADirCreate = 4, // Response to a create dir command 
	//rfbADirDelete = 5, // Response to a delete dir command 
	//rfbAFileCreate = 6, // Response to a create file command 
	rfbAFileDelete = 7, // Response to a delete file command
	
	// rfbCommand Command - content params
	rfbCDirCreate = 1, // Request the server to create the given directory
	rfbCDirDelete = 2, // Request the server to delete the given directory
	rfbCFileCreate = 3, // Request the server to create the given file
	rfbCFileDelete = 4, // Request the server to delete the given file
	
	// Errors - content params or "size" field
	rfbRErrorUnknownCmd = 1, // Unknown FileTransfer command.
	rfbRErrorCmd = 0xFFFFFFFF,
	
	// Error when a command fails on remote side (ret in "size" field)
	sz_rfbBlockSize = 8192, // new FT protocole (v2)
	
	// Size of a File Transfer packet (before compression)
	//sz_rfbZipDirectoryPrefix = 9;
	Eliminar=0;
	//String rfbZipDirectoryPrefix = "!UVNCDIR-\0";
	// Transfered directory are zipped in a file with this prefix. Must end with "-"
	// End of FileTransfer part 

	//Parsing Rfb message to see what type 
	void readRfbFileTransferMsg() 
	{
		try{
			int contentType = is.readUnsignedByte();
			int contentParamT = is.readUnsignedByte();
			int contentParam = contentParamT;
			contentParamT = is.readUnsignedByte();
			contentParamT = contentParamT << 8;
			contentParam = contentParam | contentParamT;
			if (contentType == rfbRDrivesList || contentType == rfbDirPacket)
			{
				Log.d(TAG, "rfbRDrivesList || rfbDirPacket");
				readDriveOrDirectory(contentParam);
			}
			else if (contentType == rfbFileHeader)
			{
				Log.d(TAG, "rfbFileHeader");
				receiveFileHeader();
			}
			else if (contentType == rfbFilePacket)
			{
				Log.d(TAG, "rfbFilePacket");
				receiveFileChunk();
			}
			else if (contentType == rfbEndOfFile)
			{
				Log.d(TAG, "rfbEndOfFile ok");
				endOfReceiveFile(true); // Ok
				
			}
			else if (contentType == rfbAbortFileTransfer)
			{
				Log.d(TAG, "rfbAbortFileTransfer");
				if (fFileReceptionRunning)
				{
					endOfReceiveFile(false); // Error
				}
				else
				{
					// sf@2004 - Todo: Add TestPermission 
					Log.d(TAG,"File Transfer Aborted!");
				}
				
			}
			else if (contentType == rfbCommandReturn)
			{
				Log.d(TAG, "rfbCommandReturn");
				createDirectoryorDeleteFile(contentParam);
			}
			else if (contentType == rfbFileAcceptHeader)
			{
				Log.d(TAG, "rfbFileAcceptHeader");
				sendFile();
			}
			else if (contentType == rfbFileChecksums)
			{
				Log.d(TAG, "rfbFileChecksums");
				ReceiveDestinationFileChecksums();
			}
			else
			{
				Log.d(TAG, "ContentType: " + contentType);
			}
		}catch (Exception e) {
			// TODO: handle exception
			ExploradorArchivosRemoto.semaforo.release();
			Log.d(TAG, "Exception: "+e.toString());
			
		}
	}

	//Refactored from readRfbFileTransferMsg()
	public void createDirectoryorDeleteFile(int contentParam)
		throws IOException {
		if (contentParam == rfbADirCreate)
		{
			createRemoteDirectoryFeedback();
		}
		else if (contentParam == rfbAFileDelete)
		{
			deleteRemoteFileFeedback();
		}
	}

	//Refactored from readRfbFileTransferMsg()
	public void readDriveOrDirectory(int contentParam)throws IOException{
			Log.d(TAG, "Dentro de la funcion readDriveorDirectory");
			if (contentParam == rfbADrivesList)
			{
				Log.d(TAG, "readFTPMsgDriveList");
				readFTPMsgDriveList();
			}
			else if (contentParam == rfbADirectory && !inDirectory2)
			{
				inDirectory2 = true;
				Log.d(TAG, "readFTPMsgDirectoryList");
				readFTPMsgDirectoryList();
			}
			else if (contentParam == rfbADirectory && inDirectory2)
			{
				Log.d(TAG,"readFTPMsgDirectoryListContent" );
				readFTPMsgDirectoryListContent();
			}
			else if (contentParam == 0)
			{
				Log.d(TAG, "readFTPMsgDIrectoryListEndContent");
				readFTPMsgDirectoryListEndContent();
				inDirectory2 = false;
			}
			else
			{
				Log.d(TAG, "Else: ContentParam: "+contentParam);
				//System.out.println("ContentParam: " + contentParam);
			}
		
		
	}

	// Internally used. Write an Rfb message to the server
	void writeRfbFileTransferMsg(
								int contentType,
								int contentParam,
								long size, // 0 : compression not supported - 1 : compression supported
								long length,
								String text) throws IOException
	{
		Log.d(TAG, "WriteRfbFileTransferMsg");
		byte b[] = new byte[12];

		b[0] = (byte) rfbFileTransfer;
		b[1] = (byte) contentType;
		b[2] = (byte) contentParam;

		byte by = 0;
		long c = 0;
		length++;
		c = size & 0xFF000000;
		by = (byte) (c >>> 24);
		b[4] = by;
		c = size & 0xFF0000;
		by = (byte) (c >>> 16);
		b[5] = by;
		c = size & 0xFF00;
		by = (byte) (c >>> 8);
		b[6] = by;
		c = size & 0xFF;
		by = (byte) c;
		b[7] = by;

		c = length & 0xFF000000;
		by = (byte) (c >>> 24);
		b[8] = by;
		c = length & 0xFF0000;
		by = (byte) (c >>> 16);
		b[9] = by;
		c = length & 0xFF00;
		by = (byte) (c >>> 8);
		b[10] = by;
		c = length & 0xFF;
		by = (byte) c;
		b[11] = by;
		os.write(b);
		

		if (text != null)
		{
			byte byteArray[] = text.getBytes();
			byte byteArray2[] = new byte[byteArray.length + 1];
			for (int i = 0; i < byteArray.length; i++) {
				byteArray2[i] = byteArray[i];
			}
			byteArray2[byteArray2.length - 1] = 0;
			os.write(byteArray2);
		}
		
	}

	//Internally used. Write an rfb message to the server for sending files ONLY 
	int writeRfbFileTransferMsgForSendFile(
											int contentType,
											int contentParam,
											long size,
											long length,
											String source
											) throws IOException
	{
		File f = new File(source);
		fis = new FileInputStream(f);
		byte byteBuffer[] = new byte[sz_rfbBlockSize]; 
		int bytesRead = fis.read(byteBuffer);
		long counter=0;
		boolean fError = false;
		
		// sf@ - Manage compression
		boolean fCompress = true;
		Deflater myDeflater = new Deflater();
		byte[] CompressionBuffer = new byte[sz_rfbBlockSize + 1024];
		int compressedSize = 0;
	
		while (bytesRead!=-1)
		{
				counter += bytesRead;
				myDeflater.setInput(byteBuffer, 0, bytesRead);
				myDeflater.finish();
				compressedSize = myDeflater.deflate(CompressionBuffer);
				myDeflater.reset();
				// If the compressed data is larger than the original one, we're dealing with
				// already compressed data
				if (compressedSize > bytesRead)
					fCompress = false;
				this.writeRfbFileTransferMsg(
											contentType,
											contentParam,
											(fCompress ? 1 : 0), 
											(fCompress ? compressedSize-1 : bytesRead-1),
											null
											);
				// Todo: Test write error !
				os.write(
						fCompress ? CompressionBuffer : byteBuffer,
						0,
						fCompress ? compressedSize : bytesRead
						);
				
				// Todo: test read error !
				bytesRead = fis.read(byteBuffer);
				
				// viewer.ftp.connectionStatus.setText("Sent: "+ counter + " bytes of "+ f.length() + " bytes");
				//viewer.ftp.jProgressBar.setValue((int)((counter * 100) / f.length()));
				//viewer.ftp.connectionStatus.setText(">>> Sending File: " + source + " - Size: " + f.length() + " bytes - Progress: " + ((counter * 100) / f.length()) + "%");
				
				if (fAbort == true)
				{
					Log.d(TAG, "Entro en msgforsendfile: "+ fAbort);
					fAbort = false;
					Log.d(TAG, "Entro en msg: "+ fAbort);
					fError = true;
					break;
				}
				try
				{
			        Thread.sleep(5);
			    }
				catch(InterruptedException e)
				{
			        System.err.println("Interrupted");
			    }			
				Log.d(TAG, "antes de salir: "+ fAbort);
		}
		Log.d(TAG, "Despues de salir: "+ fAbort);
		writeRfbFileTransferMsg(fError ? rfbAbortFileTransfer : rfbEndOfFile, 0, 0, 0, null);
		fis.close();
		return (fError ? -1 : 1);
	}

	//This method is internally used to send the file to the server once the server is ready
	void sendFile()
	{
		try
		{
			//viewer.ftp.disableButtons();
			int size = is.readInt();
			int length = is.readInt();
			for (int i = 0; i < length; i++)
			{
				System.out.print((char) is.readUnsignedByte());
			}
			
			int ret = writeRfbFileTransferMsgForSendFile(
															rfbFilePacket,
															0,
															0,
															0,
															sendFileSource);
			if (ret != 1)
			{
				vncCanvasActivity.pd.cancel();
				log.log(TAG, "Transferencia abortada");
				notificacion.notificar("Transferencia abortada");
				vncCanvas.enableRepaints();
			}
			else
			{
				vncCanvasActivity.pd.dismiss();
				log.log(TAG, "Archivo enviado correctamente");
				notificacion.notificar("Archivo enviado correctamente");
				vncCanvas.enableRepaints();
				Toast.makeText(vncCanvasActivity, "Archivo enviado correctamente", Toast.LENGTH_SHORT).show();
				
			}
		}
		catch (IOException e)
		{
			System.err.println(e);
		}
	}

	//Call this method to send a file from local pc to server
	void offerLocalFile(String source, String destinationPath)
	{
		try
		{
			Log.d(TAG, "OfferLocalFile");
			sendFileSource = source;
			File f = new File(source);
			// sf@2004 - Add support for huge files
			long lSize = f.length();
			int iLowSize = (int)(lSize & 0x00000000FFFFFFFF); 
			int iHighSize = (int)(lSize >> 32);
			
			String temp = destinationPath + f.getName();
			writeRfbFileTransferMsg(
									rfbFileTransferOffer,
									0,
									iLowSize, // f.length(),
									temp.length(),
									temp);
			
			// sf@2004 - Send the high part of the size			
			byte b[] = new byte[4];
			byte by = 0;
			long c = 0;
			c = iHighSize & 0xFF000000;
			by = (byte) (c >>> 24);
			b[0] = by;
			c = iHighSize & 0xFF0000;
			by = (byte) (c >>> 16);
			b[1] = by;
			c = iHighSize & 0xFF00;
			by = (byte) (c >>> 8);
			b[2] = by;
			c = iHighSize & 0xFF;
			by = (byte) c;
			b[3] = by;			
			os.write(b); 
		}
		catch (IOException e)
		{
			System.err.println(e);
		}
	}

	//Internally used.
	//Handles acknowledgement that the file has been deleted on the server
	void deleteRemoteFileFeedback() throws IOException
	{
		is.readInt();
		int length = is.readInt();
		String f = "";
		for (int i = 0; i < length; i++)
		{
			f += (char)is.readUnsignedByte();
		}
	
	}

	//Call this method to delete a file at server
	void deleteRemoteFile(String text)
	{
		try
		{
			String temp = text;
			writeRfbFileTransferMsg(
									rfbCommand,
									rfbCFileDelete,
									0,
									temp.length(),
									temp);
		}
		catch (IOException e)
		{
			System.err.println(e);
		}
	}

	//Internally used.
	// Handles acknowledgement that the directory has been created on the server
	void createRemoteDirectoryFeedback() throws IOException
	{
		is.readInt();
		int length = is.readInt();
		String f="";
		for (int i = 0; i < length; i++)
		{
			f += (char)is.readUnsignedByte();
		}
	
	}

	//Call this method to create a directory at server
	void createRemoteDirectory(String text)
	{
		try
		{
			String temp = text;
			writeRfbFileTransferMsg(
				rfbCommand,
				rfbCDirCreate,
				0,
				temp.length(),
				temp);
		}
		catch (IOException e)
		{
			System.err.println(e);
		}
	}

	//Call this method to get a file from the server
	void requestRemoteFile(String text, String localPath)
	{
		try
		{
			String temp = text;
			receivePath = localPath;
					
			writeRfbFileTransferMsg(
									rfbFileTransferRequest,
									0,
									1, // 0 : compression not supported - 1 : compression supported
									temp.length(),
									temp);
		}
		catch (IOException e)
		{
			System.err.println(e);
		}
	}

	//Internally used when transferring file from server. Here, the server sends
	//a rfb packet signalling that it is ready to send the file requested
	void receiveFileHeader() throws IOException
	{
		fFileReceptionRunning = true;
		fFileReceptionError = false;
		//viewer.ftp.disableButtons();
		int size = is.readInt();
		int length = is.readInt();
		
		String tempName = "";
		for (int i = 0; i < length; i++)
		{
			tempName += (char) is.readUnsignedByte();
		}

		// sf@2004 - Read the high part of file size (not yet in rfbFileTransferMsg for 
		// backward compatibility reasons...)
		int sizeH = is.readInt();
		long lSize = ((long)(sizeH) << 32) + size;
		
		receiveFileSize = lSize;
		//viewer.ftp.connectionStatus.setText("Received: 0 bytes of " + lSize + " bytes");
		fileSize=0;
		fileChunkCounter = 0;
		String fileName = receivePath;
		fos = new FileOutputStream(fileName);
		writeRfbFileTransferMsg(rfbFileHeader, 0, 0, 0, null);
	}

	//Internally used when transferring file from server. This method receives one chunk
	//of the file
	void receiveFileChunk() throws IOException
	{
		// sf@2004 - Size = 0 means file chunck not compressed
		int size = is.readInt();
		boolean fCompressed = (size != 0);
		int length = is.readInt();
		fileChunkCounter++;

		// sf@2004 - allocates buffers for file chunck reception and decompression 
		byte[] ReceptionBuffer = new byte[length + 32];

		// Read the incoming file data
		is.readFully(ReceptionBuffer,0, length);
		
		if (fCompressed)
		{
			int bufSize = sz_rfbBlockSize + 1024; // Todo: set a more accurate value here
			int decompressedSize = 0;
			byte[] DecompressionBuffer = new byte[bufSize];
			Inflater myInflater = new Inflater();
			myInflater.setInput(ReceptionBuffer);
			try
			{
				decompressedSize = myInflater.inflate(DecompressionBuffer);
			}
			catch (DataFormatException e)
			{
				System.err.println(e);
			}
			// Todo: check error !
			fos.write(DecompressionBuffer, 0, decompressedSize);
			fileSize += decompressedSize;
		}
		else
		{
		
			fos.write(ReceptionBuffer, 0, length);
			fileSize += length;
		}
		
		
		
		// viewer.ftp.connectionStatus.setText("Received: " + fileSize + " bytes of "+ receiveFileSize+ " bytes" );
		//viewer.ftp.jProgressBar.setValue((int)((fileSize * 100) / receiveFileSize));
		//viewer.ftp.connectionStatus.setText(">>> Receiving File: " + receivePath + " - Size: " + receiveFileSize + " bytes - Progress: " + ((fileSize * 100) / receiveFileSize) + "%");
		
		if (fAbort == true)
		{
			Log.d(TAG, "Entro en filechunk");
			fAbort = false;
			fFileReceptionError = true;
			writeRfbFileTransferMsg(rfbAbortFileTransfer, 0, 0, 0, null);
			
		}
		// sf@2004 - For old FT protocole only
		/*
		if(fileChunkCounter==10)
		{
			writeRfbFileTransferMsg(rfbFileHeader,0,0,0,null);
			fileChunkCounter=0;
		}
		*/
	}
	
	//Internally used when transferring file from server. Server signals end of file.
	void endOfReceiveFile(boolean fReceptionOk) throws IOException
	{
		int size = is.readInt();
		int length = is.readInt();
		fileSize=0;
		fos.close();
		
		vncCanvasActivity.pd.dismiss();
		vncCanvas.enableRepaints();
		//viewer.ftp.refreshLocalLocation();
		if (fReceptionOk && !fFileReceptionError)
		{
			log.log(TAG, "Archivo recibido correctamente");
			notificacion.notificar("Archivo recibido correctamente");
			
		}
		else
		{
			// sf@2004 - Delete the incomplete receieved file for now (until we use Delta Transfer)
			File f = new File(receivePath);
			f.delete();		
			log.log(TAG, "Error: Archivo no recibido");
			notificacion.notificar("Error: Archivo no recibido");
			
			
			
		}

		fFileReceptionError = false;
		fFileReceptionRunning = false;
	}

	//Call this method to read the contents of the server directory
	void readServerDirectory(String text)
	{
		Log.d(TAG, "readServerDirectory: "+text);
		try
		{
			String temp = text;
			writeRfbFileTransferMsg(
									rfbDirContentRequest,
									rfbRDirContent,
									0,
									temp.length(),
									temp);
		}
		catch (IOException e)
		{
			System.out.println("IOException in readServerDirectory(String text) in RfbProto!");
			System.out.println(e);
			System.out.println("End of exception");
		}

	}

	//Internally used to receive list of drives available on the server
	void readFTPMsgDriveList() throws IOException
	{
		Log.d(TAG, "readFTPMsgDriveList");
		String str = "";
		for (int i = 0; i < 4; i++)
		{
			is.readUnsignedByte();
		}
		int length = is.readInt();
		for (int i = 0; i < length; i++)
		{
			char temp = (char) is.readUnsignedByte();
			if (temp != '\0')
			{
				str += temp;
			}
		}
		ExploradorArchivosRemoto.listado = new String[(int)(str.length()/3)];
        for (int i=0, j=0; i<str.length(); i++){
		    
        	if(i%3==0){
           		ExploradorArchivosRemoto.listado[j] = new String(""+str.charAt(i)+":");
           		j++;
        	}
        }
		Log.d(TAG, "Libero semaforo para mostrar directorios");
		ExploradorArchivosRemoto.semaforo.release();

	}

	//Internally used to receive directory content from server
	//Here, the server marks the start of the directory listing
	void readFTPMsgDirectoryList() throws IOException
	{
		is.readInt();
		int length = is.readInt();
		if (length == 0)
		{
			inDirectory2 = false;
			readFTPMsgDirectorydriveNotReady();
			
		}
		else
		{
			// sf@2004 - New FT protocole sends remote directory name
			String str = "";
			for (int i = 0; i < length; i++)
			{
				char temp = (char) is.readUnsignedByte();
				if (temp != '\0')
				{
					str += temp;
				}
			}

		}
	}

	//Internally used to receive directory content from server
	//Here, the server sends one file/directory with it's attributes
	void readFTPMsgDirectoryListContent() throws IOException
	{
		String fileName = "", alternateFileName = "";
		byte contentType = 0;
		int contentParamT = 0;
		int contentParam = 0;
		byte temp = 0;
		int dwFileAttributes,
			nFileSizeHigh,
			nFileSizeLow,
			dwReserved0,
			dwReserved1;
		long ftCreationTime, ftLastAccessTime, ftLastWriteTime;
		char cFileName, cAlternateFileName;
		int length = 0;
		is.readInt();
		length = is.readInt();
		dwFileAttributes = is.readInt();
		length -= 4;
		ftCreationTime = is.readLong();
		length -= 8;
		ftLastAccessTime = is.readLong();
		length -= 8;
		ftLastWriteTime = is.readLong();
		length -= 8;
		nFileSizeHigh = is.readInt();
		length -= 4;
		nFileSizeLow = is.readInt();
		length -= 4;
		dwReserved0 = is.readInt();
		length -= 4;
		dwReserved1 = is.readInt();
		length -= 4;
		cFileName = (char) is.readUnsignedByte();
		length--;
		while (cFileName != '\0')
		{
			fileName += cFileName;
			cFileName = (char) is.readUnsignedByte();
			length--;
		}
		cAlternateFileName = (char) is.readByte();
		length--;
		while (length != 0)
		{
			alternateFileName += cAlternateFileName;
			cAlternateFileName = (char) is.readUnsignedByte();
			length--;
		}
		        /*
		if (dwFileAttributes == 268435456
			|| dwFileAttributes == 369098752
			|| dwFileAttributes == 285212672
			|| dwFileAttributes == 271056896
			|| dwFileAttributes == 824705024
			||	dwFileAttributes == 807927808
			|| dwFileAttributes == 371720192
			|| dwFileAttributes == 369623040
			|| dwFileAttributes == 805306368) // Fix: troessner
		*/
		//Log.d(TAG, "Llego");
		// Added Jef Fix (jdp) - check for FILE_ATTRIBUTE_DIRECTORY attribute bit
		// note that we're looking at a little-endian value in a big-endian world
		if ((dwFileAttributes & 0x10000000) == 0x10000000)
 		{
 			//fileName = " [" + fileName + "]";
 			remoteDirsList.add(fileName); // sf@2004
			Log.d(TAG, "if filename: "+fileName);
		}
		else
		{
			Log.d(TAG, "Else filename:"+fileName);
			remoteFilesList.add(" " + fileName); // sf@2004
		}
		//Log.d(TAG, "Paso");
		//a.add(fileName);
	}

	//Internally used to read directory content of server.
	//Here, server signals end of directory.
	void readFTPMsgDirectoryListEndContent() throws IOException
	{
		Log.d(TAG, "Dentro de readFTPMsgDirectoryListEndContent");
		is.readInt();
		int length = is.readInt();

		// sf@2004
		a.clear();
		
		
		ExploradorArchivosRemoto.listado = new String[remoteDirsList.size()+remoteFilesList.size()];
		int j=0;
		
		for (int i = 0; i < remoteDirsList.size(); i++){ 
			a.add(remoteDirsList.get(i));
			ExploradorArchivosRemoto.listado[j]= new String(remoteDirsList.get(i).toString());
			j++;
		}
		ExploradorArchivosRemoto.primer_archivo = j+1;
		for (int i = 0; i < remoteFilesList.size(); i++){ 
			a.add(remoteFilesList.get(i));
			ExploradorArchivosRemoto.listado[j]= new String(remoteFilesList.get(i).toString());
			j++;
		}
		Log.d(TAG, "file transfer primer_Archivo: "+ExploradorArchivosRemoto.primer_archivo);
		remoteDirsList.clear();
		remoteFilesList.clear();
		Log.d(TAG, "Libero semaforo para mostrar lista de ficheros");
		ExploradorArchivosRemoto.semaforo.release();
		/**Ya puedo elegir la carpeta donde quiero guardar el archivo local
		ahora hay que llamar a offer file con el destino deseado
		y hacerlo de forma progresiva.
		*/
	}

	
	//Internally used to signify the drive requested is not ready

	void readFTPMsgDirectorydriveNotReady() throws IOException{
			
			Log.d(TAG, "Remote Drive unavailable");
			Toast.makeText(vncCanvasActivity, "Directorio no disponible", Toast.LENGTH_SHORT).show();
		
	}

	//Call this method to request the list of drives on the server.
	void readServerDriveList()
	{
		Log.d(TAG, "READSERVERDRIVELIST");
		try
		{

			writeRfbFileTransferMsg(			rfbDirContentRequest,
												rfbRDrivesList,
												0,
												0,
												null);
		}
		catch (IOException e)
		{
			System.err.println(e);
		}
	}

	// sf@2004 - Read the destination file checksums data
	// We don't use it for now
	void ReceiveDestinationFileChecksums() throws IOException
	{
		int size = is.readInt();
		int length = is.readInt();
		
		byte[] ReceptionBuffer = new byte[length + 32];

		// Read the incoming file data
		is.readFully(ReceptionBuffer,0, length);

		/*
		String csData = "";
		for (int i = 0; i < length; i++)
		{
			csData += (char) is.readUnsignedByte();
		}
		*/
	
	}
	/**
	 * Liberamos la instancia para que se cree una nueva la proxima vez. 
	 *
	 */
	static public void finalizar() {
		Log.d(TAG, "Finalizo FileTransfer");
		instancia = null;
	}
}
