/* 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */
package android.androidVNC;

import java.util.List;

import android.content.Context;
import android.hardware.Sensor;
import android.hardware.SensorEvent;
import android.hardware.SensorEventListener;
import android.hardware.SensorManager;
import android.os.CountDownTimer;
import android.util.Log;
import android.view.Display;
import android.view.KeyEvent;
import android.view.MotionEvent;
import android.view.View;
import android.view.View.OnClickListener;
import android.widget.ImageButton;
import android.widget.RelativeLayout;
import android.widget.Toast;
import android.widget.ZoomControls;
import android.widget.RelativeLayout.LayoutParams;


/**
 * <p>(Se omiten acentos)</p>
 * <p>Modo para un control global del raton y teclado remotos. Permite panning por el escritorio.</p>
 * <p>Fichero: FullMouseMode.java.</p>
 * <p>Aadido al PFC en la version v0.1.</p>
 * <p><b>Proyecto original:</b></p>
 * 		<p>Web: http://code.google.com/p/android-vnc-viewer/</p>
 * 		<p>Release: 39.</p>
 *	 	<p>Migrada a release 60.</p>
 * @author Alberto Poblacion Navarro.
 * 
 */
class FullMouseMode implements AbstractInputHandler {
		/** Cadena para identificar la clase en el log interno*/
		private static final String TAG = "FullMouseMode";
		/** Objeto de la clase que se encarga del lienzo donde se pinta la imagen*/
		private static VncCanvas vncCanvas;
		/** Objeto de la clase que se encarga de registrar los eventos de entrada*/
		private VncCanvasActivity vncCanvasActivity;
		/** Controles de zoom */
		static ZoomControls zoom_control;
		/** Boton que controla el reset del zoom */
		static ImageButton reset;
		/** Objeto de la clase utilizada realizar escalados en la imagen remota del escrito del servidor*/
		private Zoom zoom;
		/** Variable que almacena el momento en el que se produce la pulsacion */
		private long down;
		/** Variable que almacena el momento en el que cesa una pulsacion */
		private long up;
		/** Tiempo en milisegundos para identificar si se ha tratado un click del raton */
		private final int click = 300;
		/** Tiempo en milisegundos para indentificar que se entra en el modo de arrastre */
		private final int drag = 3000;
		/** Temporizador para identificar si la pulsacion es un clic y un movimiento de arrastre */
		private CountDownTimer timer;
		/** Variable para guardar la informacion de un evento tactil anterior*/ 
		private MotionEvent anterior; 
		/** Flag para conocer si se entra en el modo de arrastre */
		private boolean dragmode = false;
		/** Variable que almacena el valor anterior de la fuerza que soportada por el acelerometro */
		private float m_totalForcePrev;
		/** Distancia entre el cursor y el borde para que se produzca un panning por la pantalla */
		final int border = 30;
		/** Cantidad de pixeles que se va a desplazar la pantalla */
		final int mover = 20;
		/** Objeto para tratar los eventos de los sensores */
		private SensorManager sensorManager;
		/** Lista que obtiene los distintos sensores que posee el terminal */
		private List<Sensor> lista_sensores;
		/** Oyente que se encarga de recibir los eventos lanzados por los sensores*/
		private SensorEventListener sensorEventListener;
		/** Flag para saber si esta activado el sensor*/
		private boolean activado = false;
		/** Alamacena las propiedades de la pantalla */
		Display display;
		
		/**
		 *  Modo para un control global del raton y el teclado remotos
		 *  @param vnccanvas instancia de la clase VncCanvas utilizada, guarda la imagen del escritorio remoto dibujada
		 *  @param vnccanvasactivity instancia de la clase VncCanvasActivity, se encarga de intercambiar informacion entre los eventos de entrada del cliente, el protocolo y vnccanvas
		 */
		FullMouseMode(VncCanvas vnccanvas, VncCanvasActivity vnccanvasactivity)
		{
			
			//Se inicializan las  variables
			vncCanvas = vnccanvas;
			vncCanvasActivity = vnccanvasactivity;
			display=vncCanvasActivity.getWindowManager().getDefaultDisplay();
			zoom= Zoom.Instancia(vncCanvasActivity, vncCanvas);
			//Modo de situacion de los widget en la pantalla: Modo relativo
			RelativeLayout tabla = new RelativeLayout(vncCanvasActivity);
			//Se configuran los parametros de la tabla: que ocupe toda la pantalla
			tabla.setLayoutParams(new LayoutParams(LayoutParams.FILL_PARENT ,LayoutParams.FILL_PARENT));
			//Configuracion de los parametros que se aplique a  los controles de zoom. Su tamao sera el necesario para ver el continido
			LayoutParams parametros = new RelativeLayout.LayoutParams(LayoutParams.WRAP_CONTENT, LayoutParams.WRAP_CONTENT);
			//Se aaden configuraciones: se alinea al fondo del padre y centrado horizontalmente
			parametros.addRule(RelativeLayout.ALIGN_PARENT_BOTTOM);
			parametros.addRule(RelativeLayout.CENTER_HORIZONTAL);
			
			//Se configuran los controles del zoom
			zoom_control = new ZoomControls(vncCanvasActivity);
			//Inicalamente no se muestran los controles de zoom
			zoom_control.hide();
			zoom_control.setLayoutParams(parametros);
			zoom_control.setOnZoomInClickListener(new OnClickListener(){
				@Override
				public void onClick(View v) {
					zoom.zoom_in();
				}
			});
			zoom_control.setOnZoomOutClickListener(new OnClickListener(){
				@Override
				public void onClick(View v) {
					zoom.zoom_out();
				}
			});
			//Se aaden los controles de zoom a la tabla
			tabla.addView(zoom_control);
			//Se crea un nuevo boton con la imagen que se encargara de resetar el zoom cuando se pulse
			reset= new ImageButton(vncCanvasActivity);
			reset.setImageResource(android.R.drawable.ic_menu_revert); 
			reset.setVisibility(View.INVISIBLE);
			reset.setOnClickListener(new OnClickListener(){
				@Override
				public void onClick(View v) {
					zoom.reset();	
				}
			});
			//Se aade el boton a la tabla
			tabla.addView(reset);
			
			//Se aade la tabla a la vista con los parametros configurados anteriormente
			vncCanvasActivity.addContentView(tabla, tabla.getLayoutParams());
			
			//Se configura un oyente de sensores para cerrar la conexin con un movimiento rapido del movil (shake)
			sensorManager = (SensorManager)vnccanvasactivity.getSystemService(Context.SENSOR_SERVICE);
			lista_sensores = sensorManager.getSensorList(Sensor.TYPE_ALL);
			sensorEventListener = new SensorEventListener(){
      	    	@Override
				public void onAccuracyChanged(Sensor sensor, int accuracy) {
      	    		//No hacer nada
      	    	}
      	    	@Override
				public void onSensorChanged(SensorEvent event) {
					if(event.sensor.getType() == Sensor.TYPE_ACCELEROMETER){
						// Oyente preparado para desconectar automaticamente la conexion se hay un movimiento brusco del telefono.
							double forceThreshHold = 1.2f;
				            double totalForce = 0.0f;
				            totalForce += Math.pow(event.values[SensorManager.DATA_X]/SensorManager.GRAVITY_EARTH, 2.0);
				            totalForce += Math.pow(event.values[SensorManager.DATA_Y]/SensorManager.GRAVITY_EARTH, 2.0);
				            totalForce += Math.pow(event.values[SensorManager.DATA_Z]/SensorManager.GRAVITY_EARTH, 2.0);
				            totalForce = Math.sqrt(totalForce);
				            if((totalForce < forceThreshHold) && (m_totalForcePrev > forceThreshHold))
				               {
				            	sensorManager.unregisterListener(sensorEventListener);
								Sensores.finalizar();
								FileTransfer.finalizar();
								Zoom.finalizar();
								Notificacion notificacion = Notificacion.Instancia(vncCanvasActivity);
								notificacion.notificar(vncCanvasActivity.getString(R.string.disconnected),vncCanvasActivity.getString(R.string.connection_closed));
								notificacion.eliminar();
								vncCanvasActivity.finish();
				               }
				            m_totalForcePrev = (float) totalForce; 	
					}
				}
			
      	    };
      	  	}
		
		/**
		 * Funcion llamada automaticamente cuando se produce una pulsacion en una tecla del teclado
		 * @param keyCode Codigo de la tecla pulsada
		 * @param evt Objeto con toda la informacion del evento que llama a la funcion 
		 * @return True si todo esta correcto, false en caso contrario.
		 */
		@Override
		public boolean onKeyDown(int keyCode, KeyEvent evt) {
			switch (keyCode) {
				case KeyEvent.KEYCODE_VOLUME_UP: //Zoom in
					zoom.zoom_in(); 
					break;
				case KeyEvent.KEYCODE_VOLUME_DOWN: //Zoom out
					zoom.zoom_out(); 
					break;
				case KeyEvent.KEYCODE_BACK: //Desconexion
					sensorManager.unregisterListener(sensorEventListener);
					Sensores.finalizar();
					FileTransfer.finalizar();
					Zoom.finalizar();
					Notificacion notificacion = Notificacion.Instancia(vncCanvasActivity);
					notificacion.notificar(vncCanvasActivity.getString(R.string.disconnected),vncCanvasActivity.getString(R.string.connection_closed));
					notificacion.eliminar();
					vncCanvasActivity.finish();
					break;
				case KeyEvent.KEYCODE_DPAD_CENTER: //Se manda el click derecho del raton en la posicion actual del raton
					KeyEvent camera = new KeyEvent(KeyEvent.ACTION_DOWN,KeyEvent.KEYCODE_CAMERA);
					vncCanvas.processLocalKeyEvent(KeyEvent.KEYCODE_CAMERA, camera);
					vncCanvas.processPointerEvent(MotionEvent.obtain(1, System.currentTimeMillis(), MotionEvent.ACTION_DOWN, vncCanvas.mouseX,vncCanvas.mouseY, 0), false);
					break;
				case KeyEvent.KEYCODE_SEARCH:
					if(!activado)
						sensorManager.registerListener(sensorEventListener, lista_sensores.get(0),SensorManager.SENSOR_DELAY_NORMAL);
					else sensorManager.unregisterListener(sensorEventListener); 
		      	    break;
				
				default:
					//Comportamiento por defecto del resto de las teclas
					return vncCanvasActivity.defaultKeyDownHandler( keyCode, evt);
				}
				return true;
			}
		/**
		 * Funcion llamada automaticamente cuando se suelta una tecla pulsada del teclado
		 * @param keyCode Codigo de la tecla
		 * @param evt Objeto con toda la informacion del evento que llama a la funcion 
		 * @return True si todo esta correcto, false en caso contrario.
		 */
		@Override
		public boolean onKeyUp(int keyCode, KeyEvent evt) {
			
			if(keyCode == KeyEvent.KEYCODE_DPAD_CENTER){ //Click derecho del raton
				vncCanvas.processPointerEvent(MotionEvent.obtain(1, System.currentTimeMillis(), MotionEvent.ACTION_UP, vncCanvas.mouseX,vncCanvas.mouseY, 0), false);
				KeyEvent camera = new KeyEvent(KeyEvent.ACTION_UP,KeyEvent.KEYCODE_CAMERA);
				vncCanvas.processLocalKeyEvent(KeyEvent.KEYCODE_CAMERA, camera);		
				}
			else{
				vncCanvasActivity.defaultKeyDownHandler( keyCode, evt);
			}
			return true;
		}
		/**
		 * Funcion llamada automaticamente cuando se produce un evento tactil
		 * @param event Objeto con toda la informacion del evento
		 */
		@Override
		public boolean onTouchEvent(MotionEvent event) {
			return fullmousetouch(event);			
		}
		/**
		 * Funcion llamada automaticamente cuando se produce un evento producido por el trackball.
		 * @pararm evt Objeto con la informacion del evento
		 */
		@Override
		public boolean onTrackballEvent(MotionEvent evt) {
			return false;
		}
		/**
		 * Devuelve una descripcion al usuario para que puede conocer en que modo se encuentra. 
		 * Se mostrara cuando se cambie de modo.
		 * @return Devuelve una cadena con la informacin.
		 */
		@Override
		public CharSequence getHandlerDescription() {
			return vncCanvasActivity.getResources().getText(R.string.input_mode_full_mouse);
			
		}
        /**
         * Devuelve un nombre usado internamente. Este nombre debe mantenerse constante durante cambios de idiomas o versiones.
         * @return Devuelve la cadena con el nombre del modo de entrada.  
         */
        @Override
        public String getName() {
                return "FULL MOUSE";
        }

	/**
	 * Procesa el evento pasado por parametro. Realizar la accion correspodiente enviando al servidor el evento correspondiente.
	 * Depende de la accion se manda un movimiento, un click o la accion de arrastrar
	 * @param event Objeto con la informacion del evento que se va a procesar.
	 * @return False en caso de error, true en caso contrario.
	 */
	public boolean fullmousetouch( MotionEvent event) 
	{ 
		//Se ajusta la posicion del evento para que sea compatible con el el Zoom
		event.setLocation((event.getX()/zoom.getZoom()), (event.getY()/zoom.getZoom()));
		switch (event.getAction()) {
		
			case MotionEvent.ACTION_DOWN:
				//Cuando se produce una pulsacion, se almacena cuando se produce y se inicia un cronometro.
				dragmode=false; 
				down= event.getEventTime();
				final MotionEvent evento = MotionEvent.obtain(event);
				anterior = MotionEvent.obtain(event);
				//Si es una pulsacion larga: modo de arrastre. 
				timer = new CountDownTimer(drag, 10){
	
				     public void onTick(long millisUntilFinished) {
				    	 //No hacer nada durante los intervalos
				     }
				     
				     public void onFinish() {
				    	 dragmode = true;
				    	 // Si se completa la cuenta atras se entra en el modo de arrastre "Draggin mode"
				    	 Toast.makeText(vncCanvasActivity, vncCanvasActivity.getString(R.string.drag)+" ON", Toast.LENGTH_SHORT).show();
				    	 //Se manda un click down ajustando la posicion de raton segun el zoom
						 evento.setAction(MotionEvent.ACTION_DOWN);
						 //Se ajusta las coordenadadas teniendo en cuenta la status bar
						 evento.offsetLocation(0, -1f * vncCanvas.getTop()/zoom.getZoom());
						 //Se ajusta las coordenadas para realizar el panning del escritorio
						 evento.offsetLocation(vncCanvasActivity.absoluteXPosition/zoom.getZoom(), vncCanvasActivity.absoluteYPosition/zoom.getZoom());
						 vncCanvas.processPointerEvent(evento, true);
					}
				  }.start();
				return true;
				
			case MotionEvent.ACTION_MOVE:
				//Si se prouce una movimiento fura de unos margenes de la pulsacion anterior, se cancela el cronometro para que no entre en el modo de arrastre
				float x = event.getX()*zoom.getZoom();
				float y = event.getY()*zoom.getZoom();
				int margen =10;
				
				if(!((x < anterior.getX()+margen)&&(x > anterior.getX()-margen))&&((y < anterior.getY()+margen)&&(y > anterior.getY()-margen)))
					timer.cancel();
				//Se procesa el movimiento para un posible panning
				return Move(event);
				
			case MotionEvent.ACTION_UP:
				//Cuando se levanta la pulsacion se guarda el tiempo y si se a tratado de una pulsacion corta se manda un clic. En caso contrario se habra entrado en drag mode
				up = event.getEventTime(); //Para conocer la duracion de la pulsacion
				timer.cancel();
				//Si es una pulsacion rapida: click de raton
				if (up - down < click)
				{
					//Mandas el gesto del click completo, cuando pulsas y cuando sueltas la tecla
					event.setAction(MotionEvent.ACTION_DOWN);
					//Se ajustan las coordenadadas teniendo en cuenta la status bar
					event.offsetLocation(0, -1f * vncCanvas.getTop()/zoom.getZoom());
					//Se ajustan las coordenadas para realizar el panning del escritorio
					event.offsetLocation(vncCanvasActivity.absoluteXPosition/zoom.getZoom(), vncCanvasActivity.absoluteYPosition/zoom.getZoom());
					//Se manda el click completo: pulsas y sueltas
					vncCanvas.processPointerEvent(event, true);
					event.setAction(MotionEvent.ACTION_UP);
					return vncCanvas.processPointerEvent(event, true);
				}
				
				if(dragmode) Toast.makeText(vncCanvasActivity, vncCanvasActivity.getString(R.string.drag)+" OFF" ,Toast.LENGTH_SHORT).show();
					
				//Si no ha sido un click del raton mandas que se levantado la tecla simplemente
				event.setAction(MotionEvent.ACTION_UP);
				//Se ajustan las coordenadas para realizar el panning del escritorio
				event.offsetLocation(0, -1f * vncCanvas.getTop()/zoom.getZoom());
				//Se ajustan las coordenadas para realizar el panning del escritorio
				event.offsetLocation(vncCanvasActivity.absoluteXPosition/zoom.getZoom(), vncCanvasActivity.absoluteYPosition/zoom.getZoom());
				return vncCanvas.processPointerEvent(event, true);
				
			}
		return false;
					
	}
	
	/**
	 * Realiza un panning por el escritorio remoto si se encuentra a una distancia menor que la indicada por la variable border.
	 * @param event Objeto con la informacin del evento.
	 * @return False si hay errores, true en caso contrario.
	 */
	
	private boolean Move(MotionEvent event) {
			//Se obtienen las cordenadas donde tienes el cursor
			float curX = event.getX()*zoom.getZoom();
			float curY = event.getY()*zoom.getZoom();
			int scrollX =0;
			int scrollY =0;
			//Si se pulsa el raton a menos distancia de "border" de uno de los extremos, se configura la posicion que se debe mover la imagen
			if (curX > (display.getWidth()-border)){
				scrollX +=mover;
			}else if (curX  < border){
				scrollX -=mover;
			}if (curY  > (display.getHeight()-border)){
				scrollY +=mover;
			}else if (curY  < border){
				scrollY -=mover;
			}
			
			//Se suman a la posicion absoluta la cantidad que no se tenga que mover	
			vncCanvasActivity.absoluteXPosition += scrollX;
			vncCanvasActivity.absoluteYPosition += scrollY;
			//Se comprueban que no se superen los margenes superior e izquierdo de la imagen
			if(vncCanvasActivity.absoluteXPosition < 0){
				scrollX=0;
				vncCanvasActivity.absoluteXPosition=0;
				vncCanvas.scrollTo(vncCanvasActivity.absoluteXPosition, vncCanvas.getScrollY());
			}
			if(vncCanvasActivity.absoluteYPosition < 0){
				scrollY=0;
				vncCanvasActivity.absoluteYPosition=0;
				vncCanvas.scrollTo(vncCanvas.getScrollX(), vncCanvasActivity.absoluteYPosition);
			}
			//Se comprueban los margenes derecho y inferior
			if(vncCanvasActivity.absoluteXPosition+ display.getWidth()> vncCanvas.getImageWidth()){
				vncCanvasActivity.absoluteXPosition=vncCanvas.getImageWidth()-(display.getWidth());
				vncCanvas.scrollTo(0, vncCanvas.getScrollY());
				vncCanvas.scrollTo(vncCanvasActivity.absoluteXPosition,vncCanvas.getScrollY());
				scrollX=0;
			}
			if(vncCanvasActivity.absoluteYPosition + display.getHeight()> vncCanvas.getImageHeight()+vncCanvas.getTop()){
				Log.d(TAG, "margen abajo");
				vncCanvasActivity.absoluteYPosition=vncCanvas.getImageHeight()-(vncCanvas.getHeight());
				vncCanvas.scrollTo(vncCanvas.getScrollX(),0);
				vncCanvas.scrollTo(vncCanvas.getScrollX(), vncCanvasActivity.absoluteYPosition);
				scrollY=0;
			}
			vncCanvas.scrollBy(scrollX, scrollY);
			//Se ajusta la posicion del evento la barra de notificaciones
			event.offsetLocation(0, -1f * vncCanvas.getTop()/zoom.getZoom());
			//Se ajusta la posicion del evento lo que nos movamos
			event.offsetLocation(vncCanvasActivity.absoluteXPosition/zoom.getZoom(),vncCanvasActivity.absoluteYPosition/zoom.getZoom());
			//Se procesa el evento sin el click pulsado, solo movimiento de raton
			vncCanvas.processPointerEvent(event, false);
			return true;
	}
}